(* $Id: godi_exec.ml 1168 2012-12-29 01:01:45Z gerd $  
 * ----------------------------------------------------------------------
 * Copyright 2008 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

open Printf

type collector_job =
  | J_none
  | J_collect of Unix.file_descr * (string -> unit) * (unit -> unit)
  | J_wait of int * (Unix.process_status -> unit)
  | J_sleep of float * (unit -> unit)
  | J_finish


class thread name =
object(self)
  (* This class implements the thread(s) for collecting stdout and stderr
     from spawned subprocesses.

     The methods starting with thread_* are private and run mostly in the
     thread. All the other methods can be invoked from outside.
   *)

  val mutable thread = None

  val mutable job = J_none
  val job_mutex = Mutex.create()
  val job_cond = Condition.create()
  val buf = String.create 4096

  initializer (
    self # thread_create()
  )

  method collect (fd : Unix.file_descr)
                 (ondata : string -> unit) 
                 (oneof : unit -> unit) =
    (* Start a new job and collect data arriving on [fd]. For every data
       chunk ondata is called (in this thread's context). When EOF or an
       error is detected, oneof is called (in this thread's context).
     *)
    self # set_job (J_collect(fd,ondata,oneof))


  method wait (pid : int) (onwait : Unix.process_status -> unit) =
    (* Waits for the process pid and calls onwait then *)
    self # set_job (J_wait(pid,onwait))


  method sleep (d : float) (onback : unit -> unit) =
    (* Sleeps for d seconds, and calls onback then. The sleep is interruptible
     *)
    self # set_job (J_sleep(d,onback))

  method interrupt_sleep() =
    Godi_win32.interrupt_sleep()

  method finish () =
    (* Finish this thread *)
    self # set_job J_finish


  method private set_job j =
    Mutex.lock job_mutex;
    assert(job = J_none);
    job <- j;
    Condition.signal job_cond;
    Mutex.unlock job_mutex;   


  method private thread_create() =
    let t = 
      Thread.create
	self#thread_loop
	() in
    thread <- Some t


  method private thread_loop() =
    let finish = ref false in
    while not !finish do
      Mutex.lock job_mutex;
      while job = J_none do
	Condition.wait job_cond job_mutex;
      done;
      ( match job with
	  | J_none -> ()
	  | J_collect(fd,ondata,oneof) ->
	      self # thread_collect fd ondata oneof
	  | J_wait(pid,onwait) ->
	      self # thread_wait pid onwait
	  | J_sleep(d,onback) ->
	      Godi_win32.sleep d;  onback()
	  | J_finish ->
	      finish := true
      );
      job <- J_none;
      Mutex.unlock job_mutex
    done


  method private thread_collect fd ondata oneof =
    let cont = ref true in
    while !cont do
      try
	let n = Unix.read fd buf 0 (String.length buf) in
	if n = 0 then (
	  cont := false;
	  oneof();
	)
	else (
	  ondata (String.sub buf 0 n)
	)
      with
	| Unix.Unix_error(Unix.EPIPE,_,_) ->
	    (* This is normal, believe it or not. *)
	    cont := false;
	    oneof()

	| e ->
	    prerr_endline("Exception reading from pipe: " ^ 
			    Printexc.to_string e);
	    cont := false;
	    oneof()
    done


  method private thread_wait pid onwait =
    let _, ps = Unix.waitpid [] pid in
    onwait ps

end


type process_status = Unix.process_status =
    WEXITED of int
  | WSIGNALED of int
  | WSTOPPED of int

type collector =
    [ `Buffer of Buffer.t
    | `Forward of (string -> unit)
    | `Stderr
    ]

type stderr_collector =
    [ collector
    | `Like_stdout
    ]

type emitter =
    [ `File of Godi_file.filename
    ]

type interactivity =
    { period : float;
      callback : unit -> bool;
    }


exception Command_not_found of string
exception Command_failure of (string * string array * process_status)
exception Command_discontinued


let stdout_thread_lz = lazy(new thread "stdout")
let stderr_thread_lz = lazy(new thread "stderr")
let wait_thread_lz = lazy(new thread "wait")
let interactivity_thread_lz = lazy(new thread "interactivity")


let collect collector s =
  match collector with
      `Buffer b -> Buffer.add_string b s
    | `Forward f ->
        ( try
            f s
          with
              error ->
		prerr_endline ("Exception in collector: " ^ 
				 Printexc.to_string error)
	)
    | `Stderr -> prerr_string s; flush stderr
;;


let kill pid =
  let _code =
    Sys.command
      (sprintf "%s -15 %d"
	 (Godi_file_helper.backslashify
	    (Godi_file.nativename
	       (Godi_file.filename "/bin/kill")))
	 (Godi_win32.get_process_id pid)) in
  ()



let pwd_re = Str.regexp "PWD=";;

let env_re = Str.regexp "^\\([^=]+\\)=";;


let make_cmdline args = (* for debugging only *)
  let maybe_quote f =
    if String.contains f ' ' || String.contains f '\"'
    then Filename.quote f
    else f in
  String.concat " " (List.map maybe_quote (Array.to_list args))


let call_cmd ?(env=[]) ?cd ?stdin ?stdout ?stderr ?interactivity (cmd,args) =
  (* We always use cmd.exe to invoke commands. This is necessary
     for Cygwin executables - at least calling these executables directly
     often results in failing Cygwin DLL startups. Maybe there is some
     left-over in the process environment so the Cygwin DLL thinks it is
     started by a Cygwin DLL (in some respect), and takes wrong approaches
     for DLL initialization. Using cmd.exe as intermediate step fixes
     these problems.
   *)

  let native_cmd =
    Godi_file_helper.backslashify
      (if not (String.contains cmd '/') && not (String.contains cmd '\\') then
	 cmd
       else
	 (Godi_file.nativename
	    (Godi_file.filename cmd))) in
  let cmd_exe =
    try Sys.getenv "COMSPEC"
    with Not_found -> "cmd.exe" in
  
  let all_args = Array.append [| cmd_exe; "/d"; "/c"; native_cmd |] args in

(*
prerr_endline ("EXEC " ^ String.concat " " (Array.to_list all_args));
prerr_endline ("CMDLINE " ^ make_cmdline all_args);
 *)

  let counter = ref 0 in
  let pipes = ref 0 in

  let action = ref None in
  let action_mutex = Mutex.create() in
  let action_todo = Condition.create() in
  let action_done = Condition.create() in

  let call f =
    (* Can be invoked from any thread *)
    Mutex.lock action_mutex;
    while !action <> None do
      Condition.wait action_done action_mutex
    done;
    action := Some f;
    Condition.signal action_todo;
    Mutex.unlock action_mutex    
  in

  let cleanup = ref [] in

  try
    let stdin_fd_cl = ref false in
    let stdin_fd =
      match (stdin : emitter option) with
	| None -> 
	    (* Unix.openfile "nul" [Unix.O_RDONLY] 0 *)
	    stdin_fd_cl := true; Unix.stdin
	| Some(`File f) ->
	    let n = Godi_file.nativename f in
	    Unix.openfile n [Unix.O_RDONLY] 0 in
    let close_stdin_fd () =
      if not !stdin_fd_cl then (Unix.close stdin_fd; stdin_fd_cl := true) in
    cleanup := close_stdin_fd :: !cleanup;

    let stdout_fd, close_stdout_fd_rd, close_stdout_fd_wr, stdout_startup =
      match (stdout : collector option) with
	| None ->
	    Unix.stdout, (fun () -> ()), (fun () -> ()), (fun () -> ())
	| Some `Stderr ->
	    Unix.stderr, (fun () -> ()), (fun () -> ()), (fun () -> ())
	| Some coll ->
	    let rd, wr = Unix.pipe() in
	    Unix.set_close_on_exec rd;
	    let stdout_thread = Lazy.force stdout_thread_lz in
	    let rd_closed = ref false in
	    let wr_closed = ref false in
	    let close_rd() =
	      if not !rd_closed then (Unix.close rd; rd_closed := true) in
	    let close_wr() =
	      if not !wr_closed then (Unix.close wr; wr_closed := true) in
	    let startup() = 
	      stdout_thread # collect 
		rd 
		(fun data -> call(fun () -> collect coll data))
		(fun () -> call(fun () -> close_rd(); decr counter; decr pipes));
	      incr counter;
	      incr pipes;
	    in
	    (wr, close_rd, close_wr, startup)
    in
    cleanup := [ close_stdout_fd_rd; close_stdout_fd_wr ] @ !cleanup;

    let stderr_fd, close_stderr_fd_rd, close_stderr_fd_wr, stderr_startup =
      match (stderr : stderr_collector option) with
	| None ->
	    Unix.stderr, (fun () -> ()), (fun () -> ()), (fun () -> ())
	| Some `Stderr ->
	    Unix.stderr, (fun () -> ()), (fun () -> ()), (fun () -> ())
	| Some `Like_stdout ->
	    stdout_fd, (fun () -> ()), (fun () -> ()), (fun () -> ())
	| Some (#collector as coll) ->
	    let rd, wr = Unix.pipe() in
	    Unix.set_close_on_exec rd;
	    let stderr_thread = Lazy.force stderr_thread_lz in
	    let rd_closed = ref false in
	    let wr_closed = ref false in
	    let close_rd() =
	      if not !rd_closed then (Unix.close rd; rd_closed := true) in
	    let close_wr() =
	      if not !wr_closed then (Unix.close wr; wr_closed := true) in
	    let startup() = 
	      stderr_thread # collect 
		rd 
		(fun data -> call(fun () -> collect coll data))
		(fun () -> call(fun () -> close_rd(); decr counter; decr pipes));
	      incr counter;
	      incr pipes
	    in
	    (wr, close_rd, close_wr, startup)
    in
    cleanup := [ close_stderr_fd_rd; close_stderr_fd_wr ] @ !cleanup;

    let wait_thread = Lazy.force wait_thread_lz in

    (* cd: temporarily change the current directory. Not elegant, but chdir
       functionality is missing in Unix.create_process although provided by
       Windows.
     *)
    let cur_pwd =
      try Unix.getenv "PWD" with Not_found -> Unix.getcwd() in
    let pwd, chdir_back =
      match cd with
	| None -> 
	    ( cur_pwd, (fun () -> ()) )
	| Some dir ->
	    (* Remember current wd, and keep it open so nobody can delete it *)
	    let curdir = Sys.getcwd() in
	    let curdir_fd = Unix.opendir curdir in
	    Godi_file.chdir dir;
	    let new_pwd = 
	      Godi_file.concat (Godi_file.filename cur_pwd) dir in
	    let back_flag = ref false in
	    ( Godi_file.slashname new_pwd,
	      (fun () -> 
		 if not !back_flag then (
		   Sys.chdir curdir; 
		   Unix.closedir curdir_fd;
		   back_flag := true
		 )
	      )
	    )
    in
    cleanup := chdir_back :: !cleanup;
  
    let env' =
      ("PWD", pwd) :: (List.filter (fun (n,_) -> n <> "PWD") env) in

    let process_env =
      let e = Unix.environment() in
      let e' =
	Array.of_list
	  ( (List.map (fun (n,v) -> n ^ "=" ^ v) env') @
	      (List.filter
		 (fun estr ->
		    if Str.string_match env_re estr 0 then (
		      let n = Str.matched_group 1 estr in
		      not(List.mem_assoc n env')
		    )
		    else true
		 )
		 (Array.to_list e))
	  ) in
      Array.sort Pervasives.compare e';  (* Windows keeps the env sorted *)
      e'
    in

    let pid = 
      Unix.create_process_env
	cmd_exe all_args process_env stdin_fd stdout_fd stderr_fd in
    (* most likely that create_process fails, if something fails *)

    (* From now on exceptions are catastrophic *)

    chdir_back();

    close_stdin_fd();
    close_stdout_fd_wr();
    close_stderr_fd_wr();
    let status = ref (Failure "No result") in
    stdout_startup();
    stderr_startup();

    (* Manage to wait for the process: *)
    let have_waiter = ref false in
    let wait_startup () =
      wait_thread # wait
	pid
	(fun ps -> 
	   call
	     (fun () -> 
		status := Command_failure(cmd, args, ps); decr counter));
      incr counter;
      have_waiter := true
    in

    (* Maybe we have interactivity: *)
    let have_ia = ref None in
    let stop_ia = ref false in
    ( match interactivity with
	| Some ia when ia.period >= 0.0 ->
	    let interactivity_thread = Lazy.force interactivity_thread_lz in
	    let rec sched_sleep() =
	      interactivity_thread # sleep ia.period 
		(fun () ->
		   call
		     (fun () ->
			let cont =
			  try ia.callback()
			  with e ->
			    prerr_endline("Exception in interactivity callback: "
					  ^ Printexc.to_string e);
			    false in
			if not cont then kill pid;
			if cont && not !stop_ia then 
			  sched_sleep() 
			else
			  decr counter
		     )
		) in
	    sched_sleep();
	    incr counter;
	    have_ia := Some interactivity_thread
	| _ -> ()
    );
    
    (* We can only wait for the process after all pipe stuff is done. Otherwise
       we'll never see the EOF of the pipes. I guess this is a Windows feature
       improving stability or so...
     *)
    if !pipes = 0 then wait_startup();

    (* Now execute all actions *)
    while !counter > 0 do
      Mutex.lock action_mutex;
      while !action = None do
	Condition.wait action_todo action_mutex
      done;
      ( match !action with
	  | None -> ()
	  | Some f -> f()
      );
      action := None;
      if !pipes = 0 && not !have_waiter then wait_startup();
      ( match !have_ia with
	  | Some iathread when !counter = 1 ->
	      stop_ia := true;
	      iathread # interrupt_sleep()
	  | _ -> ()
      );
      Condition.signal action_done;
      Mutex.unlock action_mutex;
    done;
    
    close_stdout_fd_rd();
    close_stderr_fd_rd();
  
    ( try
	raise !status
      with
	| Command_failure(_,_,Unix.WEXITED 0) ->
	    ()
	| Command_failure(c,a,Unix.WEXITED n) when n > 255 ->
	    (* Cygwin encodes WSIGNALED in the normal exit code *)
	    (* prerr_endline("SIGNAL " ^ string_of_int n); *)
	    raise (Command_failure(c,a,Unix.WSIGNALED(n lsr 8)))
	| Command_failure(c,a,Unix.WEXITED n) as e ->
	    (* prerr_endline("EXIT " ^ string_of_int n); *)
	    raise e
    )
  with
    | e ->
	List.iter (fun f -> f()) !cleanup;
	raise e
;;


(* cmd.exe has unfixable quoting problems. As workaround we write the script
   as temporary file and execute it - if the command is too complicated,
   quoting-wise
 *)

let simple_cmd_re =
  Str.regexp "^[- a-zA-Z_+/0-9'.,:;]+$"


let shell ?(setx=false) cmd =
  let simple_switches, switches =
    if setx then "-xc","-x" else "-c","" in
  if Str.string_match simple_cmd_re cmd 0 then
    ("/bin/ash", [| simple_switches; cmd |])
  else (
    let tmpdir = Godi_file.allocate_tmpdir() in
    let tmpdir_path = Godi_file.tmpdir_path tmpdir in
    let script = Godi_file.addname tmpdir_path "script" in
    let f = Godi_file.open_out `Binary script in
    output_string f cmd;
    close_out f;
    ("/bin/ash", [| switches; Godi_file.slashname script |])
      (* Files will be removed later by release_all_tmpdirs *)
  )
;;


type signal_handler_id = unit

let register_signal_handler _ _ = ()
let forget_signal_handler () = ()
let input_line() = Pervasives.input_line stdin

let environment() = Unix.environment()
let set_path s = Unix.putenv "PATH" s
let set_variable = Unix.putenv

